import * as http from 'http';
import * as https from 'https';
import { constants } from 'http2';
import { pluralize } from 'inflection';
import { BaseSchema, ObjectSchema } from 'valibot';
import { SerializerPair } from './serializers';
import {
	handleCreateItem, handleDeleteItem, handleEmplaceItem,
	handleGetCollection,
	handleGetItem,
	handleGetRoot,
	handleHasMethodAndUrl, handlePatchItem,
} from './handlers';

export interface DataSource<T = object> {
	initialize(): Promise<unknown>;
	getMultiple(): Promise<T[]>;
	getSingle(id: string): Promise<T | null>;
	create(data: Partial<T>): Promise<T>;
	delete(id: string): Promise<unknown>;
	emplace(id: string, data: Partial<T>): Promise<T>;
	patch(id: string, data: Partial<T>): Promise<T | null>;
}

export interface ApplicationParams {
	name: string;
	dataSource?: (resource: Resource) => DataSource;
}

interface ResourceFactory {
	shouldCheckSerializersOnDelete(b: boolean): this;
	shouldThrow404OnDeletingNotFound(b: boolean): this;
	id(newIdAttr: string, params: IdParams): this;
	fullText(fullTextAttr: string): this;
	name(n: string): this;
	collection(n: string): this;
	route(n: string): this;
}

export interface ResourceData {
	idAttr: string;
	itemName?: string;
	collectionName?: string;
	routeName?: string;
	newId(dataSource: DataSource): string | number | unknown;
	schema: BaseSchema;
	throws404OnDeletingNotFound: boolean;
	checksSerializersOnDelete: boolean;
}

export type Resource = ResourceData & ResourceFactory;

export interface ResourceWithDataSource extends Resource {
	dataSource: DataSource;
}

interface GenerationStrategy {
	(dataSource: DataSource, ...args: unknown[]): Promise<string | number | unknown>;
}

interface IdParams {
	generationStrategy: GenerationStrategy;
}

export const resource = <T extends BaseSchema>(schema: T): Resource => {
	let theIdAttr: string;
	let theItemName: string;
	let theCollectionName: string;
	let theRouteName: string;
	let idGenerationStrategy: GenerationStrategy;
	let throw404OnDeletingNotFound = true;
	let checkSerializersOnDelete = false;
	const fullTextAttrs = new Set<string>();

	return {
		shouldCheckSerializersOnDelete(b = true) {
			checkSerializersOnDelete = b;
			return this;
		},
		get checksSerializersOnDelete() {
			return checkSerializersOnDelete;
		},
		shouldThrow404OnDeletingNotFound(b = true) {
			throw404OnDeletingNotFound = b;
			return this;
		},
		get throws404OnDeletingNotFound() {
			return throw404OnDeletingNotFound;
		},
		id(newIdAttr: string, params: IdParams) {
			theIdAttr = newIdAttr;
			idGenerationStrategy = params.generationStrategy;
			return this;
		},
		newId(dataSource: DataSource) {
			return idGenerationStrategy(dataSource);
		},
		fullText(fullTextAttr: string) {
			if (
				schema.type === 'object'
				&& (
					schema as unknown as ObjectSchema<
						Record<string, BaseSchema>,
						undefined,
						Record<string, string>
					>
				)
					.entries[fullTextAttr]?.type === 'string'
			) {
				fullTextAttrs.add(fullTextAttr);
				return this;
			}

			throw new Error(`Could not set attribute ${fullTextAttr} as fulltext.`);
		},
		name(n: string) {
			theItemName = n;
			theCollectionName = theCollectionName ?? pluralize(theItemName).toLowerCase();
			theRouteName = theRouteName ?? theCollectionName;
			return this;
		},
		collection(n: string) {
			theCollectionName = n;
			theRouteName = theRouteName ?? theCollectionName;
			return this;
		},
		route(n: string) {
			theRouteName = n;
			return this;
		},
		get idAttr() {
			return theIdAttr;
		},
		get collectionName() {
			return theCollectionName;
		},
		get itemName() {
			return theItemName;
		},
		get routeName() {
			return theRouteName;
		},
		get schema() {
			return schema;
		}
	} as Resource;
};

interface CreateServerParams {
	baseUrl?: string;
	host?: string;
	cert?: string;
	key?: string;
	requestTimeout?: number;
}

type RequestListenerWithReturn<
	P extends unknown = unknown,
	Q extends typeof http.IncomingMessage = typeof http.IncomingMessage,
	R extends typeof http.ServerResponse = typeof http.ServerResponse
> = (
	...args: Parameters<http.RequestListener<Q, R>>
) => P;

interface HandlerState {
	handled: boolean;
}

interface ApplicationState {
	resources: Set<ResourceWithDataSource>;
	serializers: Map<string, SerializerPair>;
}

interface MiddlewareArgs {
	handlerState: HandlerState;
	appState: ApplicationState;
	appParams: ApplicationParams;
	serverParams: CreateServerParams;
}

export interface Middleware {
	(args: MiddlewareArgs): RequestListenerWithReturn<HandlerState | Promise<HandlerState>>
}

export interface Application {
	contentType(mimeTypePrefix: string, serializerPair: SerializerPair): this;
	resource(resRaw: Partial<Resource>): this;
	createServer(serverParams?: CreateServerParams): http.Server | https.Server;
}

export const application = (appParams: ApplicationParams): Application => {
	const applicationState: ApplicationState = {
		resources: new Set<ResourceWithDataSource>(),
		serializers: new Map<string, SerializerPair>()
	};

	return {
		contentType(mimeTypePrefix: string, serializerPair: SerializerPair) {
			applicationState.serializers.set(mimeTypePrefix, serializerPair);
			return this;
		},
		resource(resRaw: Partial<Resource>) {
			const res = resRaw as Partial<ResourceWithDataSource>;
			res.dataSource = res.dataSource ?? appParams.dataSource?.(res as Resource);
			if (typeof res.dataSource === 'undefined') {
				throw new Error(`Resource ${res.itemName} must have a data source.`);
			}
			applicationState.resources.add(res as ResourceWithDataSource);
			return this;
		},
		createServer(serverParams = {} as CreateServerParams) {
			const server = 'key' in serverParams && 'cert' in serverParams
				? https.createServer({
					key: serverParams.key,
					cert: serverParams.cert,
					requestTimeout: serverParams.requestTimeout
				})
				: http.createServer({
					requestTimeout: serverParams.requestTimeout
				});

			server.on('request', async (req, res) => {
				// TODO return method not allowed error when operations are not allowed on a resource
				const middlewareState = await [
					handleHasMethodAndUrl,
					handleGetRoot,
					handleGetCollection,
					handleGetItem,
					handleCreateItem,
					handleEmplaceItem,
					handlePatchItem,
					handleDeleteItem,
				]
					.reduce(
						async (currentHandlerStatePromise, middleware) => {
							const currentHandlerState = await currentHandlerStatePromise;
							if (currentHandlerState.handled) {
								return currentHandlerState;
							}

							return middleware({
								handlerState: currentHandlerState,
								appState: applicationState,
								appParams,
								serverParams
							})(req, res);
						},
						Promise.resolve<HandlerState>({
							handled: false
						})
					);

				if (middlewareState.handled) {
					return;
				}

				res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
				res.statusMessage = 'URL Not Found';
				res.end();
			});

			return server;
		}
	};
};
