import {readFile, writeFile} from 'fs/promises';
import {DataSource as DataSourceInterface, Resource} from '../core';

export class DataSource<T extends Record<string, string>> implements DataSourceInterface<T> {
	private readonly path: string;

	data: T[] = [];

	constructor(private readonly resource: Resource) {
		this.path = `${this.resource.collectionName}.jsonl`;
	}

	async initialize() {
		try {
			const fileContents = await readFile(this.path, 'utf-8');
			const lines = fileContents.split('\n');
			this.data = lines.filter((l) => l.trim().length > 0).map((l) => JSON.parse(l));
		} catch (err) {
			await writeFile(this.path, '');
		}
	}

	async getMultiple() {
		return [...this.data];
	}

	async getSingle(id: string) {
		const foundData = this.data.find((s) => s[this.resource.idAttr as string].toString() === id);

		if (foundData) {
			return {
				...foundData
			};
		}

		return null;
	}

	async create(data: Partial<T>) {
		const newData = [
			...this.data,
			data
		];

		await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));

		return data as T;
	}

	async delete(id: string) {
		const oldDataLength = this.data.length;

		const newData = this.data.filter((s) => !(s[this.resource.idAttr as string].toString() === id));

		await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));

		return oldDataLength !== newData.length;
	}

	async emplace(id: string, data: Partial<T>) {
		const existing = await this.getSingle(id);
		const dataToEmplace = {
			...data,
			[this.resource.idAttr]: id, // TODO properly serialize it to data source.
		};

		if (existing) {
			const newData = this.data.map((d) => {
				if (d[this.resource.idAttr as string].toString() === id) {
					return dataToEmplace;
				}

				return d;
			});

			await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));

			return data as T;
		}

		return this.create(dataToEmplace);
	}

	async patch(id: string, data: Partial<T>) {
		const existing = await this.getSingle(id);

		if (!existing) {
			return null;
		}

		const newItem = {
			...existing,
			...data,
		}

		const newData = this.data.map((d) => {
			if (d[this.resource.idAttr as string].toString() === id) {
				return newItem;
			}

			return d;
		});

		await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));

		return newItem as T;
	}
}
